from component.app import app
from component.app_state import app_state
from component.log import logger
from component.messagebus import consumer, publisher
from config import app_name, app_enable_message_consumer
from helper.async_task import create_task

app_state.set_liveness(True)
messages = []


async def on_error(exception: Exception):
    logger.critical(exception)
    app_state.set_readiness(False)


@app.on_event('startup')
async def startup_event():
    logger.info(f'{app_name} started')
    if app_enable_message_consumer:
        create_task(consumer.run(), on_error=on_error)
    app_state.set_readiness(True)


@consumer.register('coba')
async def handle_event(message):
    messages.append(message)
    print(messages)


@app.get('/api/get')
def handle_get():
    return messages


@app.get('/api/send')
async def handle_send():
    return await publisher.publish('coba', 'sesuatu')



from typing import Any, Optional
from core.messagebus.messagebus import (
    Publisher,  MessageSerializer, get_message_serializer
)
import aiormq
import asyncio
import logging


class RMQPublisher(Publisher):
    def __init__(
        self,
        logger: logging.Logger,
        connection_string: str,
        serializer: Optional[MessageSerializer] = None,
        retry: int = 3,
        retry_interval: int = 5
    ):
        self.logger = logger
        self.connection_string = connection_string
        self.connection: Optional[aiormq.Connection] = None
        self.serializer = get_message_serializer(serializer)
        self.retry = retry
        self.retry_interval = retry_interval

    async def publish(self, event_name: str, message: Any):
        for attempt in range(self.retry):
            try:
                if self.connection is None or self.connection.is_closed:
                    await self._connect()
                self.logger.info('🐰 Get channel')
                channel = await self.connection.channel()
                self.logger.info(f'🐰 Declare queue to publish: {event_name}')
                await channel.queue_declare(event_name)
                self.logger.info(f'🐰 Publish to "{event_name}": {message}')
                await channel.basic_publish(
                    body=self.serializer.encode(event_name, message),
                    routing_key=event_name,
                )
                return
            except Exception as e:
                self.logger.error(f'🐰 Failed to publish message: {e}')
                await self._disconnect()
                await asyncio.sleep(self.retry_interval)
                continue
        self.logger.error(
            f'🐰 Failed to publish message after {self.retry} attempts'
        )
        raise RuntimeError('Failed to publish message after retrying')

    async def _connect(self):
        self.logger.info('🐰 Create publisher connection')
        self.connection = await aiormq.connect(self.connection_string)
        self.logger.info('🐰 Publisher connection created')

    async def _disconnect(self):
        self.logger.info('🐰 Close publisher connection')
        if self.connection is not None:
            await self.connection.close()
        self.logger.info('🐰 Publisher connection closed')
        self.connection = None



from typing import Any, Callable, Mapping, Optional
from core.messagebus.messagebus import (
    Consumer, THandler, MessageSerializer, get_message_serializer
)
import asyncio
import aiormq
import inspect
import logging


class RMQConsumer(Consumer):
    def __init__(
        self,
        logger: logging.Logger,
        connection_string: str,
        serializer: Optional[MessageSerializer] = None,
        retry: int = 5,
        retry_interval: int = 5
    ):
        self.logger = logger
        self.connection_string = connection_string
        self.connection: Optional[aiormq.Connection] = None
        self.serializer = get_message_serializer(serializer)
        self.retry = retry
        self.retry_interval = retry_interval
        self._handlers: Mapping[str, THandler] = {}

    def register(self, event_name: str) -> Callable[[THandler], Any]:
        def wrapper(handler: THandler):
            self.logger.warning(f'🐰 Register handler for "{event_name}"')
            self._handlers[event_name] = handler
            return handler
        return wrapper

    async def run(self):
        return await self._run(self.retry)

    async def _run(self, retry: int):
        try:
            if self.connection is None or self.connection.is_closed:
                await self._connect()
            self.logger.info('🐰 Get channel')
            channel = await self.connection.channel()
            for event_name in self._handlers:
                self.logger.info(f'🐰 Declare queue to consume: {event_name}')
                await channel.queue_declare(event_name)
                on_message = self._create_consumer_callback(
                    channel, event_name
                )
                await channel.basic_consume(
                    queue=event_name, consumer_callback=on_message
                )
            retry = self.retry
            while True:
                await asyncio.sleep(0.01)
        except Exception:
            if retry == 0:
                self.logger.error(
                    f'🐰 Failed to consume message after {self.retry} attempts'
                )
                raise
            await self._disconnect()
            await asyncio.sleep(self.retry_interval)
            await self._run(retry-1)
        finally:
            await self._disconnect()

    async def _connect(self):
        self.logger.info('🐰 Create consumer connection')
        self.connection = await aiormq.connect(self.connection_string)
        self.logger.info('🐰 Consumer connection created')

    async def _disconnect(self):
        self.logger.info('🐰 Close consumer connection')
        if self.connection is not None:
            await self.connection.close()
        self.logger.info('🐰 Consumer connection closed')
        self.connection = None

    def _create_consumer_callback(
        self,
        channel: aiormq.Channel,
        event_name: str,
    ) -> Callable[[Any], Any]:
        async def on_message(message):
            decoded_value = self.serializer.decode(event_name, message.body)
            handler = self._handlers.get(event_name)
            self.logger.info(f'🐰 Consume from "{event_name}": {decoded_value}')
            await self._run_handler(handler, decoded_value)
            await channel.basic_ack(message.delivery_tag)
        return on_message

    async def _run_handler(
        self, message_handler: THandler, decoded_value: Any
    ):
        if inspect.iscoroutinefunction(message_handler):
            return asyncio.create_task(message_handler(decoded_value))
        return message_handler(decoded_value)



from config import (
    app_name, app_broker_type, app_rmq_connection_string, app_kafka_bootstrap_servers
)
from core.messagebus.mock import MockConsumer, MockPublisher
from core.messagebus.messagebus import Consumer, MessageSerializer, Publisher
from core.messagebus.rabbitmq.consumer import RMQConsumer
from core.messagebus.kafka.consumer import KafkaConsumer
from core.messagebus.kafka.publisher import KafkaPublisher
from core.messagebus.rabbitmq.publisher import RMQPublisher
from component.log import logger


def init_publisher(
    broker_type: str,
    serializer: MessageSerializer,
    default_publisher: Publisher
) -> Publisher:
    if broker_type == 'rabbitmq':
        return RMQPublisher(
            logger=logger,
            connection_string=app_rmq_connection_string,
            serializer=serializer
        )
    if broker_type == 'kafka':
        return KafkaPublisher(
            logger=logger,
            bootstrap_servers=app_kafka_bootstrap_servers,
            serializer=serializer
        )
    if broker_type == 'mock':
        return default_publisher
    raise Exception(f'Invalid broker type: {broker_type}')


def init_consumer(
    broker_type: str,
    serializer: MessageSerializer,
    default_consumer: Consumer
) -> Consumer:
    if broker_type == 'rabbitmq':
        return RMQConsumer(
            logger=logger,
            connection_string=app_rmq_connection_string,
            serializer=serializer
        )
    if broker_type == 'kafka':
        return KafkaConsumer(
            logger=logger,
            bootstrap_servers=app_kafka_bootstrap_servers,
            group_id=app_name,
            serializer=serializer
        )
    if broker_type == 'mock':
        return default_consumer
    raise Exception(f'Invalid broker type: {broker_type}')


def init_message_serializer() -> MessageSerializer:
    # Add custom logic if necessary
    return MessageSerializer()


message_serializer = init_message_serializer()
mock_consumer = MockConsumer(logger, message_serializer)
mock_publisher = MockPublisher(logger, mock_consumer, message_serializer)
consumer = init_consumer(app_broker_type, message_serializer, mock_consumer)
publisher = init_publisher(app_broker_type, message_serializer, mock_publisher)



from typing import Any, Callable, Mapping, Optional
from core.messagebus.messagebus import (
    Consumer, THandler, MessageSerializer,
    get_message_serializer
)
from aiokafka import AIOKafkaConsumer, __version__
from aiokafka.consumer.consumer import RoundRobinPartitionAssignor

import asyncio
import inspect
import logging


class KafkaConsumer(Consumer):
    def __init__(
        self,
        logger: logging.Logger,
        bootstrap_servers: str,
        client_id='aiokafka-' + __version__,
        group_id: Optional[str] = None,
        key_deserializer=None,
        value_deserializer=None,
        fetch_max_wait_ms=500,
        fetch_max_bytes=52428800,
        fetch_min_bytes=1,
        max_partition_fetch_bytes=1 * 1024 * 1024,
        request_timeout_ms=40 * 1000,
        retry_backoff_ms=100,
        auto_offset_reset='latest',
        enable_auto_commit=True,
        auto_commit_interval_ms=5000,
        check_crcs=True,
        metadata_max_age_ms=5 * 60 * 1000,
        partition_assignment_strategy=(RoundRobinPartitionAssignor,),
        max_poll_interval_ms=300000,
        rebalance_timeout_ms=None,
        session_timeout_ms=10000,
        heartbeat_interval_ms=3000,
        consumer_timeout_ms=200,
        max_poll_records=None,
        ssl_context=None,
        security_protocol='PLAINTEXT',
        api_version='auto',
        exclude_internal_topics=True,
        connections_max_idle_ms=540000,
        isolation_level="read_uncommitted",
        sasl_mechanism="PLAIN",
        sasl_plain_password=None,
        sasl_plain_username=None,
        sasl_kerberos_service_name='kafka',
        sasl_kerberos_domain_name=None,
        sasl_oauth_token_provider=None,
        serializer: Optional[MessageSerializer] = None,
        retry: int = 5,
        retry_interval: int = 5
    ):
        self.logger = logger
        self.serializer = get_message_serializer(serializer)
        self.bootstrap_servers = bootstrap_servers
        self.client_id = client_id
        self.group_id = group_id
        self.key_deserializer = key_deserializer
        self.value_deserializer = value_deserializer
        self.fetch_max_wait_ms = fetch_max_wait_ms
        self.fetch_max_bytes = fetch_max_bytes
        self.fetch_min_bytes = fetch_min_bytes
        self.max_partition_fetch_bytes = max_partition_fetch_bytes
        self.request_timeout_ms = request_timeout_ms
        self.retry_backoff_ms = retry_backoff_ms
        self.auto_offset_reset = auto_offset_reset
        self.enable_auto_commit = enable_auto_commit
        self.auto_commit_interval_ms = auto_commit_interval_ms
        self.check_crcs = check_crcs
        self.metadata_max_age_ms = metadata_max_age_ms
        self.partition_assignment_strategy = partition_assignment_strategy
        self.max_poll_interval_ms = max_poll_interval_ms
        self.rebalance_timeout_ms = rebalance_timeout_ms
        self.session_timeout_ms = session_timeout_ms
        self.heartbeat_interval_ms = heartbeat_interval_ms
        self.consumer_timeout_ms = consumer_timeout_ms
        self.max_poll_records = max_poll_records
        self.ssl_context = ssl_context
        self.security_protocol = security_protocol
        self.api_version = api_version
        self.exclude_internal_topics = exclude_internal_topics
        self.connections_max_idle_ms = connections_max_idle_ms
        self.isolation_level = isolation_level
        self.sasl_mechanism = sasl_mechanism
        self.sasl_plain_password = sasl_plain_password
        self.sasl_plain_username = sasl_plain_username
        self.sasl_kerberos_service_name = sasl_kerberos_service_name
        self.sasl_kerberos_domain_name = sasl_kerberos_domain_name
        self.sasl_oauth_token_provider = sasl_oauth_token_provider
        self.consumer: Optional[AIOKafkaConsumer] = None
        self.retry = retry
        self.retry_interval = retry_interval
        self._handlers: Mapping[str, THandler] = {}

    def register(self, event_name: str) -> Callable[[THandler], Any]:
        def wrapper(handler: THandler):
            self.logger.warning(f'🐼 Register handler for "{event_name}"')
            self._handlers[event_name] = handler
            return handler
        return wrapper

    async def run(self):
        return await self._run(self.retry)

    async def _run(self, retry: int):
        try:
            if self.consumer is None:
                await self._connect()
            topics = list(self._handlers.keys())
            self.logger.warning(f'🐼 Subscribe to topics: {topics}')
            self.consumer.subscribe(topics=topics)
            async for message in self.consumer:
                event_name = message.topic
                message_handler = self._handlers.get(event_name)
                decoded_value = self.serializer.decode(
                    event_name, message.value
                )
                self.logger.info(
                    f'🐼 Consume "{event_name}": {decoded_value}'
                )
                await self._run_handler(message_handler, decoded_value)
            retry = self.retry
        except Exception:
            if retry == 0:
                self.logger.error(
                    f'🐼 Failed to consume message after {self.retry} attempts'
                )
                self.logger.fatal('🐼 Cannot retry')
                raise
            self.logger.warning('🐼 Retry to consume')
            await self._disconnect()
            await asyncio.sleep(self.retry_interval)
            await self._run(retry-1)
        finally:
            await self._disconnect()

    async def _connect(self):
        self.logger.info('🐼 Create kafka consumer')
        self.consumer = AIOKafkaConsumer(
            bootstrap_servers=self.bootstrap_servers,
            client_id=self.client_id,
            group_id=self.group_id,
            key_deserializer=self.key_deserializer,
            value_deserializer=self.value_deserializer,
            fetch_max_wait_ms=self.fetch_max_wait_ms,
            fetch_max_bytes=self.fetch_max_bytes,
            fetch_min_bytes=self.fetch_min_bytes,
            max_partition_fetch_bytes=self.max_partition_fetch_bytes,
            request_timeout_ms=self.request_timeout_ms,
            retry_backoff_ms=self.retry_backoff_ms,
            auto_offset_reset=self.auto_offset_reset,
            enable_auto_commit=self.enable_auto_commit,
            auto_commit_interval_ms=self.auto_commit_interval_ms,
            check_crcs=self.check_crcs,
            metadata_max_age_ms=self.metadata_max_age_ms,
            partition_assignment_strategy=self.partition_assignment_strategy,
            max_poll_interval_ms=self.max_poll_interval_ms,
            rebalance_timeout_ms=self.rebalance_timeout_ms,
            session_timeout_ms=self.session_timeout_ms,
            heartbeat_interval_ms=self.heartbeat_interval_ms,
            consumer_timeout_ms=self.consumer_timeout_ms,
            max_poll_records=self.max_poll_records,
            ssl_context=self.ssl_context,
            security_protocol=self.security_protocol,
            api_version=self.api_version,
            exclude_internal_topics=self.exclude_internal_topics,
            connections_max_idle_ms=self.connections_max_idle_ms,
            isolation_level=self.isolation_level,
            sasl_mechanism=self.sasl_mechanism,
            sasl_plain_password=self.sasl_plain_password,
            sasl_plain_username=self.sasl_plain_username,
            sasl_kerberos_service_name=self.sasl_kerberos_service_name,
            sasl_kerberos_domain_name=self.sasl_kerberos_domain_name,
            sasl_oauth_token_provider=self.sasl_oauth_token_provider,
        )
        self.logger.info('🐼 Start kafka consumer')
        await self.consumer.start()
        self.logger.info('🐼 Kafka consumer started')

    async def _disconnect(self):
        if self.consumer is not None:
            self.logger.info('🐼 Unsubscribe kafka consumer from all topics')
            self.consumer.unsubscribe()
            self.logger.info('🐼 Stop kafka consumer')
            await self.consumer.stop()
            self.logger.info('🐼 Kafka consumer stopped')
        self.consumer = None

    async def _run_handler(
        self, message_handler: THandler, decoded_value: Any
    ):
        if inspect.iscoroutinefunction(message_handler):
            return asyncio.create_task(message_handler(decoded_value))
        return message_handler(decoded_value)



from zrb import CmdTask, DockerComposeTask, EnvFile, runner
from zrb.builtin._group import project_group
from ._common import (
    CURRENT_DIR, APP_DIR, RESOURCE_DIR, TEMPLATE_ENV_FILE_NAME,
    SKIP_SUPPORT_CONTAINER_EXECUTION, SKIP_LOCAL_MONOLITH_EXECUTION,
    SKIP_LOCAL_MICROSERVICES_EXECUTION,
    rabbitmq_checker, rabbitmq_management_checker,
    redpanda_console_checker, kafka_outside_checker,
    kafka_plaintext_checker, pandaproxy_outside_checker,
    pandaproxy_plaintext_checker, app_local_checker,
    local_input, mode_input, host_input, https_input, image_input,
    local_app_port_env, local_app_broker_type_env, compose_app_host_port_env,
    compose_rabbitmq_host_port_env, compose_rabbitmq_management_host_port_env,
    compose_redpanda_console_host_port_env,
    compose_kafka_outside_host_port_env,
    compose_pandaproxy_outside_host_port_env,
    compose_kafka_plaintext_host_port_env,
    compose_pandaproxy_plaintext_host_port_env,
    start_support_container_compose_profile_env,
)
from .image import build_snake_app_name_image
from .frontend import build_snake_app_name_frontend
from .container import remove_snake_app_name_container
import os

support_compose_env_prefix = 'CONTAINER_ENV_PREFIX'
support_compose_envs = [
    local_app_broker_type_env,
    local_app_port_env,
    compose_app_host_port_env,
    compose_rabbitmq_host_port_env,
    compose_rabbitmq_management_host_port_env,
    compose_redpanda_console_host_port_env,
    compose_kafka_outside_host_port_env,
    compose_kafka_plaintext_host_port_env,
    compose_pandaproxy_outside_host_port_env,
    compose_pandaproxy_plaintext_host_port_env
]
app_env_file = EnvFile(env_file=TEMPLATE_ENV_FILE_NAME, prefix='ENV_PREFIX')
app_envs = [
    local_app_broker_type_env,
    local_app_port_env,
]

###############################################################################
# Task Definitions
###############################################################################

start_snake_app_name_support_container = DockerComposeTask(
    icon='🐳',
    name='start-kebab-app-name-support-container',
    description='Start human readable app name container',
    inputs=[
        local_input,
        host_input,
        https_input,
        image_input,
    ],
    skip_execution=SKIP_SUPPORT_CONTAINER_EXECUTION,
    upstreams=[
        build_snake_app_name_image,
        remove_snake_app_name_container
    ],
    cwd=RESOURCE_DIR,
    compose_cmd='up',
    compose_env_prefix=support_compose_env_prefix,
    envs=support_compose_envs + [
        start_support_container_compose_profile_env,
    ],
    checkers=[
        rabbitmq_checker,
        rabbitmq_management_checker,
        kafka_outside_checker,
        kafka_plaintext_checker,
        redpanda_console_checker,
        pandaproxy_outside_checker,
        pandaproxy_plaintext_checker,
    ]
)

start_monolith_snake_app_name = CmdTask(
    icon='🚤',
    name='start-monolith-kebab-app-name',
    inputs=[
        local_input,
        mode_input,
        host_input,
        https_input
    ],
    skip_execution=SKIP_LOCAL_MONOLITH_EXECUTION,
    upstreams=[
        start_snake_app_name_support_container,
        build_snake_app_name_frontend
    ],
    cwd=APP_DIR,
    env_files=[app_env_file],
    envs=app_envs,
    cmd_path=os.path.join(CURRENT_DIR, 'cmd', 'start.sh'),
    checkers=[
        app_local_checker,
    ]
)

start_snake_app_name_gateway = CmdTask(
    icon='🚤',
    name='start-kebab-app-name-gateway',
    inputs=[
        local_input,
        mode_input,
        host_input,
        https_input
    ],
    skip_execution=SKIP_LOCAL_MICROSERVICES_EXECUTION,
    upstreams=[
        start_snake_app_name_support_container,
        build_snake_app_name_frontend
    ],
    cwd=APP_DIR,
    env_files=[app_env_file],
    envs=app_envs,
    cmd_path=os.path.join(CURRENT_DIR, 'cmd', 'start.sh'),
    checkers=[
        app_local_checker,
    ]
)

start_snake_app_name = CmdTask(
    icon='🚤',
    name='start-kebab-app-name',
    description='Start human readable app name',
    group=project_group,
    cmd='echo "Ok"',
    upstreams=[
        start_monolith_snake_app_name,
        start_snake_app_name_gateway,
    ]
)
runner.register(start_snake_app_name)
